@@ -1,29 +1,38 @@
 import ast
 import atexit
 import concurrent.futures
-import importlib.resources
+import contextlib
 import linecache
-import math
 import re
 import shutil
-import subprocess  # nosec
 import sys
-from dataclasses import dataclass
 from inspect import signature
 from pathlib import Path
 from tempfile import TemporaryDirectory
-from threading import Lock
+from textwrap import dedent
 from types import ModuleType
-from typing import Callable, Optional, TextIO, Tuple, TypeVar, Union, overload
+from typing import Any, Callable, Optional, TextIO, TypeVar, Union, overload
 
-from autoregistry import Registry
-from pathspec import PathSpec
 from serial import SerialException
 from serial.tools.miniterm import Miniterm
 from typing_extensions import ParamSpec
 
 from ._minify import minify as minify_code
-from .exceptions import ConnectionLost, MaxHistoryLengthError
+from .device_meta import DeviceMeta
+from .device_support import Implementation, MethodMetadata, sort_executers
+from .device_sync_support import (
+    discover_files_dirs,
+    generate_dst_dirs,
+    preprocess_ignore,
+    preprocess_keep,
+    preprocess_src_file_hash,
+)
+from .exceptions import (
+    ConnectionLost,
+    InternalError,
+    MaxHistoryLengthError,
+    NotBelayResponseError,
+)
 from .executers import (
     Executer,
     SetupExecuter,
@@ -31,30 +40,67 @@ from .executers import (
     TeardownExecuter,
     ThreadExecuter,
 )
-from .hash import fnv1a
-from .helpers import read_snippet, wraps_partial
+from .helpers import get_fnv1a32_native_path, read_snippet, wraps_partial
 from .inspect import isexpression
+from .proxy_object import ProxyObject
 from .pyboard import Pyboard, PyboardError, PyboardException
 from .typing import BelayReturn, PathType
 from .webrepl import WebreplToSerial
 
+if sys.version_info < (3, 9, 0):
+    import importlib_resources
+else:
+    import importlib.resources as importlib_resources
+
 P = ParamSpec("P")
 R = TypeVar("R")
 
 
-class NotBelayResponse(Exception):
-    """Parsed response wasn't for Belay."""
+def remove_call(expression: str, func_name: str, required=False) -> str:
+    expression = expression.strip()
+    func_name_with_parenthesis = func_name + "("
+    if expression.startswith(func_name_with_parenthesis) and expression[-1] == ")":
+        return expression[len(func_name_with_parenthesis) : -1]
+
+    if required:
+        raise ValueError
+
+    return expression
 
 
-def _parse_belay_response(line):
+def extended_literal_eval(expression):
+    """
+    Extended version of ast.literal_eval that also handles other common builtins.
+    """
+    # If it's a string that looks like dict_keys, convert it to a list
+    if isinstance(expression, str) and expression.startswith("dict_keys("):
+        return list(ast.literal_eval(remove_call(expression, "dict_keys")))
+
+    return ast.literal_eval(expression)
+
+
+def parse_belay_response(
+    line: str,
+    result_parser: Callable[[str], Any] = extended_literal_eval,
+):
+    """Parse a Belay response string into a python object.
+
+    Parameters
+    ----------
+    line: str
+        String representations of a python object.
+        e.g. "(1, 2, 'foo')"
+    result_parser: Callable
+        Function that accepts a string and returns a python object.
+    """
     if not line.startswith("_BELAY"):
-        raise NotBelayResponse
+        raise NotBelayResponseError
     line = line[6:]
     code, line = line[0], line[1:]
 
     if code == "R":
         # Result
-        return ast.literal_eval(line)
+        return result_parser(line)
     elif code == "S":
         # StopIteration
         raise StopIteration
@@ -62,182 +108,12 @@ def _parse_belay_response(line):
         raise ValueError(f'Received unknown code: "{code}"')
 
 
-def _discover_files_dirs(
-    remote_dir: str,
-    local_file_or_folder: Path,
-    ignore: Optional[list] = None,
-):
-    src_objects = []
-    if local_file_or_folder.is_dir():
-        if ignore is None:
-            ignore = []
-        ignore_spec = PathSpec.from_lines("gitwildmatch", ignore)
-        for src_object in local_file_or_folder.rglob("*"):
-            if ignore_spec.match_file(str(src_object)):
-                continue
-            src_objects.append(src_object)
-        # Sort so that folder creation comes before file sending.
-        src_objects.sort()
-
-        src_files, src_dirs = [], []
-        for src_object in src_objects:
-            if src_object.is_dir():
-                src_dirs.append(src_object)
-            else:
-                src_files.append(src_object)
-        dst_files = [
-            remote_dir / src.relative_to(local_file_or_folder) for src in src_files
-        ]
-    else:
-        src_files = [local_file_or_folder]
-        src_dirs = []
-        dst_files = [Path(remote_dir) / local_file_or_folder.name]
-
-    return src_files, src_dirs, dst_files
-
-
-def _preprocess_keep(
-    keep: Union[None, list, str, bool],
-    dst: str,
-) -> list:
-    if keep is None:
-        if dst == "/":
-            keep = ["boot.py", "webrepl_cfg.py", "lib"]
-        else:
-            keep = []
-    elif isinstance(keep, str):
-        keep = [keep]
-    elif isinstance(keep, (list, tuple)):
-        pass
-    elif isinstance(keep, bool):
-        keep = []
-    else:
-        raise ValueError
-    keep = [(dst / Path(x)).as_posix() for x in keep]
-    return keep
-
-
-def _preprocess_ignore(ignore: Union[None, str, list, tuple]) -> list:
-    if ignore is None:
-        ignore = ["*.pyc", "__pycache__", ".DS_Store", ".pytest_cache"]
-    elif isinstance(ignore, str):
-        ignore = [ignore]
-    elif isinstance(ignore, (list, tuple)):
-        ignore = list(ignore)
-    else:
-        raise ValueError
-    return ignore
-
-
-def _preprocess_src_file(
-    tmp_dir: PathType,
-    src_file: PathType,
-    minify: bool,
-    mpy_cross_binary: Union[str, Path, None],
-) -> Path:
-    tmp_dir = Path(tmp_dir)
-    src_file = Path(src_file)
-
-    if src_file.is_absolute():
-        transformed = tmp_dir / src_file.relative_to(tmp_dir.anchor)
-    else:
-        transformed = tmp_dir / src_file
-    transformed.parent.mkdir(parents=True, exist_ok=True)
-
-    if src_file.suffix == ".py":
-        if mpy_cross_binary:
-            transformed = transformed.with_suffix(".mpy")
-            subprocess.check_output(  # nosec
-                [mpy_cross_binary, "-o", transformed, src_file]
-            )
-            return transformed
-        elif minify:
-            minified = minify_code(src_file.read_text())
-            transformed.write_text(minified)
-            return transformed
-
-    return src_file
-
-
-def _preprocess_src_file_hash(*args, **kwargs):
-    src_file = _preprocess_src_file(*args, **kwargs)
-    src_hash = fnv1a(src_file)
-    return src_file, src_hash
-
-
-def _generate_dst_dirs(dst, src, src_dirs) -> list:
-    dst_dirs = [(dst / x.relative_to(src)).as_posix() for x in src_dirs]
-    # Add all directories leading up to ``dst``.
-    dst_prefix_tokens = dst.split("/")
-    for i in range(2, len(dst_prefix_tokens) + (dst[-1] != "/")):
-        dst_dirs.append("/".join(dst_prefix_tokens[:i]))
-    dst_dirs.sort()
-    return dst_dirs
-
-
-def _sort_executers(executers):
-    """Sorts executers by monotonically increasing ``__belay__.id``.
-
-    This ensures that, when necessary, executers are called in the order they are defined.
-    """
-
-    def get_key(x):
-        try:
-            return x.__wrapped__.__belay__.id
-        except AttributeError:
-            return math.inf
-
-    return sorted(executers, key=get_key)
-
-
-@dataclass
-class Implementation:
-    """Implementation dataclass detailing the device.
-
-    Parameters
-    ----------
-    name: str
-        Type of python running on device.
-        One of ``{"micropython", "circuitpython"}``.
-    version: Tuple[int, int, int]
-        ``(major, minor, patch)`` Semantic versioning of device's firmware.
-    platform: str
-        Board identifier. May not be consistent from MicroPython to CircuitPython.
-        e.g. The Pi Pico is "rp2" in MicroPython, but "RP2040"  in CircuitPython.
-    emitters: tuple[str]
-        Tuple of available emitters on-device ``{"native", "viper"}``.
-    """
-
-    name: str
-    version: Tuple[int, int, int]
-    platform: str
-    emitters: Tuple[str]
-
-
-_method_metadata_counter_lock = Lock()
-_method_metadata_counter = 0
-
-
-@dataclass
-class MethodMetadata:
-    executer: Callable
-    kwargs: dict
-    autoinit: bool = False  # Only applies to ``SetupExecuter``.
-    id: int = -1  # monotonically increasing global identifier.
-
-    def __post_init__(self):
-        global _method_metadata_counter
-        with _method_metadata_counter_lock:
-            self.id = _method_metadata_counter
-            _method_metadata_counter += 1
-
-
-class Device(Registry):
+class Device(metaclass=DeviceMeta):
     """Belay interface into a micropython device.
 
     Can be used as a context manager; calls ``self.close`` on exit.
 
-    Inherits from ``autoregistry.Registry`` for easy access to subclasses.
+    Uses the ``autoregistry.RegistryMeta`` metaclass for easy-to-access subclasses.
 
     Attributes
     ----------
@@ -270,43 +146,56 @@ class Device(Registry):
 
         self._connect_to_board(**self._board_kwargs)
 
-        for executer_name, executer_cls in Executer.items():
-            executer = executer_cls(self)
-            setattr(
-                self, executer_name, executer
-            )  # Public interface; might get stomped
-            setattr(
-                self, "_belay_" + executer_name, executer
-            )  # Private interface; will always be there
-
         self._exec_snippet("startup")
 
+        # Obtain implementation early on so implementation-specific executers can be bound.
         self.implementation = Implementation(
             *self(
-                "(sys.implementation.name, sys.implementation.version, sys.platform)"
+                "(sys.implementation.name, sys.implementation.version, sys.platform, getattr(sys.implementation, '_mpy', None))"
             ),
             emitters=self._emitter_check(),
         )
 
-        # if subclassing Device, register methods decorated with
-        # executer markers (e.g. ``Device.task``).
+        # Setup executer generators and bind to private attributes.
+        executer_generators = {}
+        for executer_name, executer_cls in Executer.items():
+            executer_generator = executer_cls(self)
+            # Private interface, will always be available
+            attr_name = f"_belay_{executer_name}"
+            setattr(self, attr_name, executer_generator)
+            executer_generators[executer_name] = executer_generator
+
+        # If subclassing Device, register methods decorated with
+        # executer markers (e.g. ``@Device.task``).
         autoinit_executers = []
-        for name, method in vars(type(self)).items():
-            metadata = getattr(method, "__belay__", None)
-            if not metadata:
+        instantiated_executer_names = set()
+        for method_name in dir(type(self)):
+            # Get method from self to trigger descriptors.
+            try:
+                method = getattr(self, method_name)
+                metadata = method.__belay__
+            except AttributeError:
                 continue
-            decorator = getattr(self, metadata.executer.__registry__.name)
-            executer = decorator(method, **metadata.kwargs)
+            executer_name = metadata.executer.__registry__.name
+            executer_generator = executer_generators[executer_name]
+            executer = executer_generator(method, **metadata.kwargs)
+            instantiated_executer_names.add(executer_name)
 
             if metadata.autoinit:
                 autoinit_executers.append(executer)
 
             setattr(
                 self,
-                name,
+                method_name,
                 executer,
             )
 
+        # Setup publicly accessible if the name hasn't been stomped.
+        for executer_name, executer_generator in executer_generators.items():
+            if executer_name in instantiated_executer_names:
+                continue
+            setattr(self, executer_name, executer_generator)
+
         if startup is None:
             if self.implementation.name == "circuitpython":
                 self._exec_snippet("convenience_imports_circuitpython")
@@ -317,8 +206,7 @@ class Device(Registry):
 
         self.__pre_autoinit__()
 
-        autoinit_executers = _sort_executers(autoinit_executers)
-        for executer in autoinit_executers:
+        for executer in sort_executers(autoinit_executers):
             executer()
 
         atexit.register(self.close)
@@ -346,7 +234,7 @@ class Device(Registry):
             self._exec_snippet("emitter_check")
         except PyboardException as e:
             if "invalid micropython decorator" not in str(e):
-                raise e
+                raise
             # Get line of exception
             line_e = int(re.findall(r"line (\d+)", str(e))[-1])
             if line_e == 1:
@@ -358,7 +246,7 @@ class Device(Registry):
                     # viper is not available
                     pass
                 else:
-                    raise Exception(f"Unknown emitter line {line_e}.")
+                    raise InternalError(f"Unknown emitter line {line_e}.") from e
         else:
             emitters.append("native")
             emitters.append("viper")
@@ -368,10 +256,7 @@ class Device(Registry):
 
     def _connect_to_board(self, **kwargs):
         self._board = Pyboard(**kwargs)
-        if isinstance(self._board.serial, WebreplToSerial):
-            soft_reset = False
-        else:
-            soft_reset = True
+        soft_reset = not isinstance(self._board.serial, WebreplToSerial)
         self._board.enter_raw_repl(soft_reset=soft_reset)
 
     def _exec_snippet(self, *names: str) -> BelayReturn:
@@ -392,6 +277,7 @@ class Device(Registry):
         minify: bool = True,
         stream_out: TextIO = sys.stdout,
         record=True,
+        trusted: bool = False,
     ):
         """Execute code on-device.
 
@@ -406,11 +292,20 @@ class Device(Registry):
         record: bool
             Record the call for state-reconstruction if device is accidentally reset.
             Defaults to ``True``.
+        trusted: bool
+            Fully trust remote device.
+            When set to ``False``, only ``[None, bool, bytes, int, float, str, List, Dict, Set]`` return
+            values can be parsed.
+            When set to ``True``, any value who's ``repr`` can be evaluated to create a python object can be
+            returned. However, **this also allows the remote device to execute arbitrary code on host**.
+            Defaults to ``False``.
 
         Returns
         -------
             Correctly interpreted return value from executing code on-device.
         """
+        cmd = dedent(cmd)
+
         if minify:
             cmd = minify_code(cmd)
 
@@ -418,44 +313,54 @@ class Device(Registry):
             # Belay Tasks are inherently expressions as well.
             cmd = f"print('_BELAYR' + repr({cmd}))"
 
-        if (
-            record
-            and self.attempts
-            and len(self._cmd_history) < self.MAX_CMD_HISTORY_LEN
-        ):
+        if record and self.attempts and len(self._cmd_history) < self.MAX_CMD_HISTORY_LEN:
             self._cmd_history.append(cmd)
 
-        out = None
-        data_consumer_buffer = []
+        out = None  # Used to store the parsed response object.
+        data_consumer_buffer = bytearray()
 
         def data_consumer(data):
             """Handle input data stream immediately."""
             nonlocal out
-            if data == b"\x04":
+            data = data.replace(b"\x04", b"")
+            if not data:
                 return
-            data_consumer_buffer.append(data.decode())
-            if b"\n" in data:
-                line = "".join(data_consumer_buffer)
-                data_consumer_buffer.clear()
-
+            data_consumer_buffer.extend(data)
+            while (i := data_consumer_buffer.find(b"\n")) >= 0:
+                i += 1
+                line = data_consumer_buffer[:i].decode()
+                data_consumer_buffer[:] = data_consumer_buffer[i:]
                 try:
-                    out = _parse_belay_response(line)
-                except NotBelayResponse:
+                    if trusted:
+                        out = parse_belay_response(line, result_parser=eval)
+                    else:
+                        out = parse_belay_response(line)
+                except NotBelayResponseError:
                     if stream_out:
                         stream_out.write(line)
 
         try:
             self._board.exec(cmd, data_consumer=data_consumer)
-        except (SerialException, ConnectionResetError):
+        except (SerialException, ConnectionResetError) as e:
             # Board probably disconnected.
             if self.attempts:
                 self.reconnect()
-                self._board.exec(cmd, data_consumer=data_consumer_buffer)
+                self._board.exec(cmd, data_consumer=data_consumer)
             else:
-                raise ConnectionLost
+                raise ConnectionLost from e
 
         return out
 
+    def proxy(self, name: str) -> ProxyObject:
+        """Create a :class:`.ProxyObject` that uses this :class:`Device`.
+
+        Parameters
+        ----------
+        name: str
+            Name of the remote object for the proxy-object to interact with.
+        """
+        return ProxyObject(self, name)
+
     def sync(
         self,
         folder: PathType,
@@ -508,66 +413,95 @@ class Device(Registry):
         if not folder.exists():
             raise ValueError(f'"{folder}" does not exist.')
 
-        # Create a list of all files and dirs (on-device).
-        # This is so we know what to clean up after done syncing.
-        snippets_to_execute = []
-
-        if progress_update:
-            progress_update(description="Bootstrapping sync...")
-        if "viper" in self.implementation.emitters:
-            snippets_to_execute.append("hf_viper")
-        elif "native" in self.implementation.emitters:
-            snippets_to_execute.append("hf_native")
-        else:
-            snippets_to_execute.append("hf")
-
-        if self.implementation.name == "circuitpython":
-            snippets_to_execute.append("ilistdir_circuitpython")
-        else:
-            snippets_to_execute.append("ilistdir_micropython")
-        snippets_to_execute.append("sync_begin")
-        self._exec_snippet(*snippets_to_execute)
-
-        # Remove the keep files from the on-device ``all_files`` set
-        # so they don't get deleted.
-        keep_all = folder.is_file() or keep is True
-        keep = _preprocess_keep(keep, dst)
-        ignore = _preprocess_ignore(ignore)
-
-        src_files, src_dirs, dst_files = _discover_files_dirs(dst, folder, ignore)
-
-        if mpy_cross_binary:
-            dst_files = [
-                dst_file.with_suffix(".mpy") if dst_file.suffix == ".py" else dst_file
-                for dst_file in dst_files
-            ]
-        dst_files = [dst_file.as_posix() for dst_file in dst_files]
-        dst_dirs = _generate_dst_dirs(dst, folder, src_dirs)
-
-        if keep_all:
-            self("del __belay_del_fs")
-        else:
-            self(
-                f"__belay_del_fs({repr(dst)}, {repr(set(keep + dst_files))}); del __belay_del_fs"
-            )
-
-        # Try and make all remote dirs
-        if dst_dirs:
-            if progress_update:
-                progress_update(description="Creating remote directories...")
-            self(f"__belay_mkdirs({repr(dst_dirs)})")
-
         with TemporaryDirectory() as tmp_dir, concurrent.futures.ThreadPoolExecutor() as executor:
             tmp_dir = Path(tmp_dir)
 
+            # Create a list of all files and dirs (on-device).
+            # This is so we know what to clean up after done syncing.
+            snippets_to_execute = []
+
+            # Remove the keep files from the on-device ``all_files`` set
+            # so they don't get deleted.
+            keep_all = folder.is_file() or keep is True
+            keep = preprocess_keep(keep, dst)
+            ignore = preprocess_ignore(ignore)
+
+            src_files, src_dirs, dst_files = discover_files_dirs(dst, folder, ignore)
+
             def _preprocess_src_file_hash_helper(src_file):
-                return _preprocess_src_file_hash(
-                    tmp_dir, src_file, minify, mpy_cross_binary
+                return preprocess_src_file_hash(tmp_dir, src_file, minify, mpy_cross_binary)
+
+            futures = [executor.submit(_preprocess_src_file_hash_helper, src_file) for src_file in src_files]
+
+            if progress_update:
+                progress_update(description="Bootstrapping sync...")
+
+            if self.implementation.name == "circuitpython":
+                snippets_to_execute.append("ilistdir_circuitpython")
+            else:
+                snippets_to_execute.append("ilistdir_micropython")
+
+            snippets_to_execute.append("sync_begin")
+
+            # Place the fnv1a32 at the end of snippets_to_execute, as there's a chance
+            # the micropython interpreter wasn't compiled with all the required features.
+            fnv1a32_native_path = get_fnv1a32_native_path(self.implementation)
+            if fnv1a32_native_path:
+                # Sync a pre-compiled native fnv1a32 hashing module.
+                src_size = fnv1a32_native_path.stat().st_size
+                try:
+                    # Check if file exists and has the same size
+                    remote_size = self("os.stat('/_belay_fnv1a32.mpy')[6]")
+                    sync_nativemodule = remote_size != src_size
+                except PyboardException:
+                    sync_nativemodule = True
+
+                if sync_nativemodule:
+                    self._board.fs_put(str(fnv1a32_native_path), "/_belay_fnv1a32.mpy")
+
+                snippets_to_execute.append("hf_nativemodule")
+            elif "viper" in self.implementation.emitters:
+                snippets_to_execute.append("hf_viper")
+            elif "native" in self.implementation.emitters:
+                snippets_to_execute.append("hf_native")
+            else:
+                snippets_to_execute.append("hf")
+
+            try:
+                self._exec_snippet(*snippets_to_execute)
+            except PyboardException:
+                if not fnv1a32_native_path:
+                    raise
+
+                # A bit fragile, but fragile for the sake of performance.
+                # This is most likely the micropython interpreter being unable to import the fnv1a32 mpy file.
+                # Fallback to one of the other fnv1a32 implementations.
+                if "viper" in self.implementation.emitters:
+                    self._exec_snippet("hf_viper")
+                elif "native" in self.implementation.emitters:
+                    self._exec_snippet("hf_native")
+                else:
+                    self._exec_snippet("hf")
+
+            if mpy_cross_binary:
+                dst_files = [
+                    dst_file.with_suffix(".mpy") if dst_file.suffix == ".py" else dst_file for dst_file in dst_files
+                ]
+            dst_files = [dst_file.as_posix() for dst_file in dst_files]
+            dst_dirs = generate_dst_dirs(dst, folder, src_dirs)
+
+            if keep_all:
+                self("del __belay_del_fs")
+            else:
+                self(
+                    f"__belay_del_fs({repr(dst)}, {repr(set(keep + dst_files + ['/_belay_fnv1a32.mpy']))}); del __belay_del_fs"
                 )
 
-            src_files_and_hashes = executor.map(
-                _preprocess_src_file_hash_helper, src_files
-            )
+            # Try and make all remote dirs
+            if dst_dirs:
+                if progress_update:
+                    progress_update(description="Creating remote directories...")
+                self(f"__belay_mkdirs({repr(dst_dirs)})")
 
             # Get all remote hashes
             if progress_update:
@@ -575,12 +509,11 @@ class Device(Registry):
             dst_hashes = self(f"__belay_hfs({repr(dst_files)})")
 
             if len(dst_hashes) != len(dst_files):
-                raise Exception
+                raise InternalError
 
+            src_files_and_hashes = [future.result() for future in futures]
             puts = []
-            for (src_file, src_hash), dst_file, dst_hash in zip(
-                src_files_and_hashes, dst_files, dst_hashes
-            ):
+            for (src_file, src_hash), dst_file, dst_hash in zip(src_files_and_hashes, dst_files, dst_hashes):
                 if src_hash != dst_hash:
                     puts.append((src_file, dst_file))
 
@@ -642,11 +575,11 @@ class Device(Registry):
         **kwargs
             Passed along to ``Device.sync``.
         """
-        pkg_files = importlib.resources.files(package)
+        pkg_files = importlib_resources.files(package)
         with TemporaryDirectory() as tmp_dir:
             tmp_dir = Path(tmp_dir)
             for subfolder in subfolders:
-                with importlib.resources.as_file(pkg_files / str(subfolder)) as f:
+                with importlib_resources.as_file(pkg_files / str(subfolder)) as f:
                     shutil.copytree(f, tmp_dir, dirs_exist_ok=True)
             self.sync(tmp_dir, dst=dst, **kwargs)
 
@@ -670,7 +603,7 @@ class Device(Registry):
 
         self._board.cancel_running_program()
 
-        for executer in _sort_executers(self._belay_teardown._belay_executers):
+        for executer in sort_executers(self._belay_teardown._belay_executers):
             executer()
 
         self._board.close()
@@ -705,19 +638,20 @@ class Device(Registry):
 
     @overload
     @staticmethod
-    def setup(f: Callable[P, R]) -> Callable[P, R]:
-        ...
+    def setup(f: Callable[P, R]) -> Callable[P, R]: ...
 
     @overload
     @staticmethod
     def setup(
-        *, autoinit=False, **kwargs
-    ) -> Callable[[Callable[P, R]], Callable[P, R]]:
-        ...
+        *, autoinit: bool = False, implementation: str = "", **kwargs
+    ) -> Callable[[Callable[P, R]], Callable[P, R]]: ...
 
     @staticmethod
     def setup(
-        f: Optional[Callable[P, R]] = None, autoinit=False, **kwargs
+        f: Optional[Callable[P, R]] = None,
+        autoinit: bool = False,
+        implementation: str = "",
+        **kwargs,
     ) -> Union[Callable[[Callable[P, R]], Callable[P, R]], Callable[P, R]]:
         """Execute decorated function's body in a global-context on-device when called.
 
@@ -742,34 +676,39 @@ class Device(Registry):
             Automatically invokes decorated functions at the end of object ``__init__``.
             Methods will be executed in order-registered.
             Defaults to ``False``.
+        ignore_errors: bool
+            Discard any device-side uncaught exception.
+            Defaults to ``False``.
+        implementation: str
+            If supplied, the provided method will **only** be used if the board's implementation **name** matches.
+            Several methods of the same name can be overloaded that support different implementations.
+            Common values include "micropython", and "circuitpython".
+            Defaults to an empty string, which **all** implementations will match to.
         """  # noqa: D400
         if f is None:
-            return wraps_partial(Device.setup, autoinit=autoinit, **kwargs)  # type: ignore[reportGeneralTypeIssues]
+            return wraps_partial(Device.setup, autoinit=autoinit, implementation=implementation, **kwargs)  # type: ignore[reportGeneralTypeIssues]
         if signature(f).parameters and autoinit:
-            raise ValueError(
-                f"Method {f} decorated with "
-                '"@Device.setup(autoinit=True)" '
-                "must have no arguments."
-            )
+            raise ValueError(f"Method {f} decorated with @Device.setup(autoinit=True) must have no arguments.")
 
         f.__belay__ = MethodMetadata(
-            executer=SetupExecuter, autoinit=autoinit, kwargs=kwargs
+            executer=SetupExecuter,
+            autoinit=autoinit,
+            implementation=implementation,
+            kwargs=kwargs,
         )
         return f
 
     @overload
     @staticmethod
-    def teardown(f: Callable[P, R]) -> Callable[P, R]:
-        ...
+    def teardown(f: Callable[P, R]) -> Callable[P, R]: ...
 
     @overload
     @staticmethod
-    def teardown(**kwargs) -> Callable[[Callable[P, R]], Callable[P, R]]:
-        ...
+    def teardown(*, implementation: str = "", **kwargs) -> Callable[[Callable[P, R]], Callable[P, R]]: ...
 
     @staticmethod
     def teardown(
-        f: Optional[Callable[P, R]] = None, **kwargs
+        f: Optional[Callable[P, R]] = None, implementation: str = "", **kwargs
     ) -> Union[Callable[[Callable[P, R]], Callable[P, R]], Callable[P, R]]:
         """Executes decorated function's body in a global-context on-device when ``device.close()`` is called.
 
@@ -790,31 +729,35 @@ class Device(Registry):
         record: bool
             Each invocation of the executer is recorded for playback upon reconnect.
             Defaults to ``True``.
+        ignore_errors: bool
+            Discard any device-side uncaught exception.
+            Defaults to ``False``.
+        implementation: str
+            If supplied, the provided method will **only** be used if the board's implementation **name** matches.
+            Several methods of the same name can be overloaded that support different implementations.
+            Common values include "micropython", and "circuitpython".
+            Defaults to an empty string, which **all** implementations will match to.
         """  # noqa: D400
         if f is None:
-            return wraps_partial(Device.teardown, **kwargs)  # type: ignore[reportGeneralTypeIssues]
+            return wraps_partial(Device.teardown, implementation=implementation, **kwargs)  # type: ignore[reportGeneralTypeIssues]
 
         if signature(f).parameters:
-            raise ValueError(
-                f'Method {f} decorated with "@Device.teardown" must have no arguments.'
-            )
+            raise ValueError(f'Method {f} decorated with "@Device.teardown" must have no arguments.')
 
-        f.__belay__ = MethodMetadata(executer=TeardownExecuter, kwargs=kwargs)
+        f.__belay__ = MethodMetadata(executer=TeardownExecuter, implementation=implementation, kwargs=kwargs)
         return f
 
     @overload
     @staticmethod
-    def task(f: Callable[P, R]) -> Callable[P, R]:
-        ...
+    def task(f: Callable[P, R]) -> Callable[P, R]: ...
 
     @overload
     @staticmethod
-    def task(**kwargs) -> Callable[[Callable[P, R]], Callable[P, R]]:
-        ...
+    def task(*, implementation: str = "", **kwargs) -> Callable[[Callable[P, R]], Callable[P, R]]: ...
 
     @staticmethod
     def task(
-        f: Optional[Callable[P, R]] = None, **kwargs
+        f: Optional[Callable[P, R]] = None, implementation: str = "", **kwargs
     ) -> Union[Callable[[Callable[P, R]], Callable[P, R]], Callable[P, R]]:
         """Execute decorated function on-device.
 
@@ -837,27 +780,36 @@ class Device(Registry):
             Each invocation of the executer is recorded for playback upon reconnect.
             Only recommended to be set to ``True`` for a setup-like function.
             Defaults to ``False``.
+        implementation: str
+            If supplied, the provided method will **only** be used if the board's implementation **name** matches.
+            Several methods of the same name can be overloaded that support different implementations.
+            Common values include "micropython", and "circuitpython".
+            Defaults to an empty string, which **all** implementations will match to.
+        trusted: bool
+            Fully trust remote device.
+            When set to ``False``, only ``[None, bool, bytes, int, float, str, List, Dict, Set]`` return
+            values can be parsed.
+            When set to ``True``, any value who's ``repr`` can be evaluated to create a python object can be
+            returned. However, **this also allows the remote device to execute arbitrary code on host**.
+            Defaults to ``False``.
         """  # noqa: D400
         if f is None:
-            return wraps_partial(Device.task, **kwargs)  # type: ignore[reportGeneralTypeIssues]
+            return wraps_partial(Device.task, implementation=implementation, **kwargs)  # type: ignore[reportGeneralTypeIssues]
 
-        f.__belay__ = MethodMetadata(executer=TaskExecuter, kwargs=kwargs)
+        f.__belay__ = MethodMetadata(executer=TaskExecuter, implementation=implementation, kwargs=kwargs)
         return f
 
-    @staticmethod
     @overload
     @staticmethod
-    def thread(f: Callable[P, R]) -> Callable[P, R]:
-        ...
+    def thread(f: Callable[P, R]) -> Callable[P, R]: ...
 
     @overload
     @staticmethod
-    def thread(**kwargs) -> Callable[[Callable[P, R]], Callable[P, R]]:
-        ...
+    def thread(*, implementation: str = "", **kwargs) -> Callable[[Callable[P, R]], Callable[P, R]]: ...
 
     @staticmethod
     def thread(
-        f: Optional[Callable[P, R]] = None, **kwargs
+        f: Optional[Callable[P, R]] = None, implementation: str = "", **kwargs
     ) -> Union[Callable[[Callable[P, R]], Callable[P, R]], Callable[P, R]]:
         """Spawn on-device thread that executes decorated function.
 
@@ -876,10 +828,15 @@ class Device(Registry):
         record: bool
             Each invocation of the executer is recorded for playback upon reconnect.
             Defaults to ``True``.
+        implementation: str
+            If supplied, the provided method will **only** be used if the board's implementation **name** matches.
+            Several methods of the same name can be overloaded that support different implementations.
+            Common values include "micropython", and "circuitpython".
+            Defaults to an empty string, which **all** implementations will match to.
         """  # noqa: D400
         if f is None:
-            return wraps_partial(Device.task, **kwargs)  # type: ignore[reportGeneralTypeIssues]
-        f.__belay__ = MethodMetadata(executer=ThreadExecuter, kwargs=kwargs)
+            return wraps_partial(Device.task, implementation=implementation, **kwargs)  # type: ignore[reportGeneralTypeIssues]
+        f.__belay__ = MethodMetadata(executer=ThreadExecuter, implementation=implementation, kwargs=kwargs)
         return f
 
     def terminal(self, *, exit_char=chr(0x1D)):
@@ -890,10 +847,8 @@ class Device(Registry):
         miniterm.set_tx_encoding("UTF-8")
         miniterm.exit_character = exit_char
         miniterm.start()
-        try:
+        with contextlib.suppress(KeyboardInterrupt):
             miniterm.join(True)
-        except KeyboardInterrupt:
-            pass
         miniterm.join()
 
     def soft_reset(self):
@@ -901,7 +856,7 @@ class Device(Registry):
         # When in Raw REPL, ctrl-d will perform a reset, but won't execute ``main.py``
         # https://github.com/micropython/micropython/issues/2249
         self._board.exit_raw_repl()
-        self._board.read_until(1, b">>>")
+        self._board.read_until(b">>>")
         self._board.ctrl_d()
 
     def _traceback_execute(
@@ -911,6 +866,7 @@ class Device(Registry):
         name: str,
         cmd: str,
         record: bool = True,
+        trusted: bool = False,
     ):
         """Invoke ``cmd``, and reinterprets raised stacktrace in ``PyboardException``.
 
@@ -927,6 +883,13 @@ class Device(Registry):
         record: bool
             Record the call for state-reconstruction if device is accidentally reset.
             Defaults to ``True``.
+        trusted: bool
+            Fully trust remote device.
+            When set to ``False``, only ``[None, bool, bytes, int, float, str, List, Dict, Set]`` return
+            values can be parsed.
+            When set to ``True``, any value who's ``repr`` can be evaluated to create a python object can be
+            returned. However, **this also allows the remote device to execute arbitrary code on host**.
+            Defaults to ``False``.
 
         Returns
         -------
@@ -935,7 +898,7 @@ class Device(Registry):
         src_file = str(src_file)
 
         try:
-            res = self(cmd, record=record)
+            res = self(cmd, record=record, trusted=trusted)
         except PyboardException as e:
             new_lines = []
 
